/* n = number of rows = number of columns */
%let n = 10;
data matrix;
   input c1-c&n;
   datalines;
 7 51 52 87 38 60 74 66  0 20
50 12  0 64  8 53  0 46 76 42
27 77  0 18 22 48 44 13  0 57
62  0  3  8  5  6 14  0 26 39
 0 97  0  5 13  0 41 31 62 48
79 68  0  0 15 12 17 47 35 43
76 99 48 27 34  0  0  0 28  0
 0 20  9 27 46 15 84 19  3 24
56 10 45 39  0 93 67 79 19 38
27  0 39 53 46 24 69 46 23  1
;

/* find optimal solution by calling PROC OPTNET */
proc optnet data_matrix=matrix;
   lap out=out;
run;

/* find optimal solution by calling PROC OPTNET with sparse input */
data links(keep=from to weight);
   array c[&n];
   set matrix;
   from = compress('row'||put(_N_,best.));
   do j = 1 to &n;
      to = compress('col'||put(j,best.));
      weight = c[j];
      if weight ne . then output;
   end;
run;
proc optnet data_links=links graph_direction=directed;
   lap out=out;
run;

/* find optimal solution by calling network solver in PROC OPTMODEL */
proc optmodel;
   num n = &n;
   set NSET = 1..n;
   num c {NSET, NSET};
   read data matrix into [_n_] {j in NSET} <c[_n_,j]=col('c'||j)>;

   set ARCS = setof {i in NSET, j in NSET} <'row'||i,'col'||j>;
   num cost {ARCS};
   for {i in NSET, j in NSET} cost['row'||i,'col'||j] = c[i,j];
   set <str,str> ASSIGNMENTS;

   /* call network solver */
   solve with network /
      LAP direction=directed links=(weight=cost) out=(assignments=ASSIGNMENTS);
   put ASSIGNMENTS=;
quit;

/* find optimal solution by calling MILP or LP solver in PROC OPTMODEL */
proc optmodel;
   num n = &n;
   set NSET = 1..n;
   num c {NSET, NSET};
   read data matrix into [_n_] {j in NSET} <c[_n_,j]=col('c'||j)>;

   /* X[i,j] = 1 iff i assigned to j */
   var X {NSET, NSET} binary;
   min Z = sum {i in NSET, j in NSET} c[i,j] * X[i,j];
   con Assign_i {i in NSET}:
      sum {j in NSET} X[i,j] = 1;
   con Assign_j {j in NSET}:
      sum {i in NSET} X[i,j] = 1;

   set SUPPORT = {i in NSET, j in NSET: X[i,j].sol > 0.5};

   /* call MILP solver */
   solve;
   put SUPPORT=;

   /* call LP solver */
   solve with LP relaxint;
   put SUPPORT=;
quit;

/* find optimal solution by calling CLP solver in PROC OPTMODEL */
proc optmodel;
   num n = &n;
   set NSET = 1..n;
   num c {NSET, NSET};
   read data matrix into [_n_] {j in NSET} <c[_n_,j]=col('c'||j)>;

   /* X[i,j] = 1 iff i assigned to j */
   var X {NSET, NSET} binary;
   min Z = sum {i in NSET, j in NSET} c[i,j] * X[i,j];
   con Assign_i {i in NSET}:
      sum {j in NSET} X[i,j] = 1;
   con Assign_j {j in NSET}:
      sum {i in NSET} X[i,j] = 1;

   set SUPPORT = {i in NSET, j in NSET: X[i,j].sol > 0.5};

   solve with CLP;
   put SUPPORT=;
quit;

/* find optimal solution by calling CLP solver using predicates in PROC OPTMODEL */
proc optmodel;
   num n = &n;
   set NSET = 1..n;
   num c {NSET, NSET};
   read data matrix into [_n_] {j in NSET} <c[_n_,j]=col('c'||j)>;

   /* Assign[i] = j iff i assigned to j */
   var Assign {NSET} >= 1 <= n integer;
   con AlldiffCon:
      alldiff(Assign);

   /* Cost[i] = c[i,Assign[i]] */
   var Cost {NSET} integer;
   con ElementCon {i in NSET}:
      element(Assign[i], {j in NSET} c[i,j], Cost[i]);
   min Z = sum {i in NSET} Cost[i];

   set SUPPORT = {i in NSET, j in NSET: Assign[i].sol = j};

   /* call CLP solver */
   solve;
   put SUPPORT=;
quit;

/* k = number of solutions desired */
%let k = 4;

/* find (up to) k optimal solutions by solving one CLP */
proc optmodel;
   num n = &n;
   set NSET = 1..n;
   num c {NSET, NSET};
   read data matrix into [_n_] {j in NSET} <c[_n_,j]=col('c'||j)>;

   /* Assign[i] = j iff i assigned to j */
   var Assign {NSET} >= 1 <= n integer;
   con AlldiffCon:
      alldiff(Assign);

   /* Cost[i] = c[i,Assign[i]] */
   var Cost {NSET} integer;
   con ElementCon {i in NSET}:
      element(Assign[i], {j in NSET} c[i,j], Cost[i]);
   min Z = sum {i in NSET} Cost[i];

   num k = &k;
   set SOLS = 1.._NSOL_;
   set SUPPORT_ALL = {i in NSET, j in NSET, s in SOLS: Assign[i].sol[s] = j};
   solve with CLP / maxsolns=(k);
   print Assign;
   create data soldata(drop=s) from [s]=SOLS
      cost=(sum {<i,j,(s)> in SUPPORT_ALL} c[i,j])
      {i in NSET} <col('row'||i)=Assign[i].sol[s]>;
quit;

/* find k best solutions by solving one CLP */
proc optmodel;
   num k = &k;
   set SOLS = 1..k;
   num n = &n;
   set NSET = 1..n;
   num c {NSET, NSET};
   read data matrix into [_n_] {j in NSET} <c[_n_,j]=col('c'||j)>;

   /* Assign[i,s] = j iff i assigned to j in solution s */
   var Assign {NSET, SOLS} >= 1 <= n integer;
   con AlldiffCon {s in SOLS}:
      alldiff({i in NSET} Assign[i,s]);

   /* Cost[i,s] = c[i,Assign[i,s]] */
   var Cost {NSET, SOLS} integer;
   con ElementCon {i in NSET, s in SOLS}:
      element(Assign[i,s], {j in NSET} c[i,j], Cost[i,s]);
   min Z = sum {i in NSET, s in SOLS} Cost[i,s];

   /* use LEXICO LT to get distinct solutions */
   con LexicoCon {s in SOLS diff {k}}:
      lexico({i in NSET} Assign[i,s] < {i in NSET} Assign[i,s+1]);

   solve;
   print Assign;
   create data soldata(drop=s) from [s]=SOLS
      cost=(sum {i in NSET} Cost[i,s])
      {i in NSET} <col('row'||i)=Assign[i,s]>;
quit;
proc print;
run;

/* find k best solutions by solving one MILP */
proc optmodel;
   num k = &k;
   set SOLS = 1..k;
   set SOLPAIRS = {s1 in SOLS, s2 in SOLS: s1 < s2};
   num n = &n;
   set NSET = 1..n;
   num c {NSET, NSET};
   read data matrix into [_n_] {j in NSET} <c[_n_,j]=col('c'||j)>;

   /* X[i,j,s] = 1 iff i assigned to j in solution s */
   var X {NSET, NSET, SOLS} binary;
   min Z = sum {i in NSET, j in NSET, s in SOLS} c[i,j] * X[i,j,s];
   con Assign_i {i in NSET, s in SOLS}:
      sum {j in NSET} X[i,j,s] = 1;
   con Assign_j {j in NSET, s in SOLS}:
      sum {i in NSET} X[i,j,s] = 1;

   /* Y[i,j,s1,s2] = 1 implies X[i,j,s1] ne X[i,j,s2] */
   var Y {NSET, NSET, SOLPAIRS} binary;
   /* solutions must differ in at least one component */
*   con AtLeastOneDiff {<s1,s2> in SOLPAIRS}:
      sum {i in NSET, j in NSET} Y[i,j,s1,s2] >= 1;
   /* all feasible solutions have the same number of ones,
      so can use stronger constraint */
   con AtLeastTwoDiff {<s1,s2> in SOLPAIRS}:
      sum {i in NSET, j in NSET} Y[i,j,s1,s2] >= 2;
   /* if Y[i,j,s1,s2] = 1 then not(X[i,j,s1] = 1 and X[i,j,s2] = 1) */
   con Ycon1 {i in NSET, j in NSET, <s1,s2> in SOLPAIRS}:
      X[i,j,s1] + X[i,j,s2] + Y[i,j,s1,s2] <= 2;
   /* if Y[i,j,s1,s2] = 1 then not(X[i,j,s1] = 0 and X[i,j,s2] = 0) */
   con Ycon0 {i in NSET, j in NSET, <s1,s2> in SOLPAIRS}:
      Y[i,j,s1,s2] <= X[i,j,s1] + X[i,j,s2];

   set SUPPORT_ALL = {i in NSET, j in NSET, s in SOLS: X[i,j,s].sol > 0.5};
   num assign {NSET, SOLS};

   solve with MILP / logfreq=1000;
   for {<i,j,s> in SUPPORT_ALL} assign[i,s] = j;
   put SUPPORT_ALL=;
   print assign;
   num cost {s in SOLS} = sum {i in NSET, j in NSET} c[i,j] * X[i,j,s].sol;
   print cost;
quit;

/* find k best solutions by solving k MILPs (row generation) */
proc optmodel printlevel=0;
   num n = &n;
   set NSET = 1..n;
   num c {NSET, NSET};
   read data matrix into [_n_] {j in NSET} <c[_n_,j]=col('c'||j)>;

   /* X[i,j] = 1 iff i assigned to j */
   var X {NSET, NSET} binary;
   min Z = sum {i in NSET, j in NSET} c[i,j] * X[i,j];
   con Assign_i {i in NSET}:
      sum {j in NSET} X[i,j] = 1;
   con Assign_j {j in NSET}:
      sum {i in NSET} X[i,j] = 1;

   set SUPPORT = {i in NSET, j in NSET: X[i,j].sol > 0.5};

   num k = &k;
   num numsols init 0;
   set SOLS = 1..numsols;

   set <num,num> SUPPORT_s {SOLS};
*   con Exclude {s in SOLS}:
      sum {<i,j> in SUPPORT[s]} (1 - X[i,j])
    + sum {<i,j> in NSET cross NSET diff SUPPORT[s]} X[i,j] >= 1;
   /* all feasible solutions have the same number of ones,
      so can use stronger (and sparser) cut */
   con Exclude {s in SOLS}:
      sum {<i,j> in SUPPORT_s[s]} (1 - X[i,j]) >= 1;

   for {1..k} do;
      solve with MILP / loglevel=0;
      numsols = numsols + 1;
      SUPPORT_s[numsols] = {i in NSET, j in NSET: X[i,j].sol > 0.5};
      put _OBJ_.sol SUPPORT_s[numsols]=;
   end;
quit;

/* find k best solutions by solving multiple LAPs,
   with concurrent (n-1)-way branching (Murty 1968) */
proc optmodel printlevel=0 prefertkfunc=never;
   num n = &n;
   num k = &k;
   num numsols init 0;
   set SOLS = 1..numsols;
   set <num,num> SUPPORT_s {SOLS};
   set NSET = 1..n;
   num c {NSET, NSET};
   read data matrix into [_n_] {j in NSET} <c[_n_,j]=col('c'||j)>;

   set ARCS = setof {i in NSET, j in NSET} <'row'||i,'col'||j>;
   num cost {ARCS};
   for {i in NSET, j in NSET} cost['row'||i,'col'||j] = c[i,j];
   set <str,str> ASSIGNMENTS;

   num numnodes init 1;
   set NODES = 1..numnodes; /* root is node 1 */
   /* FIX[node,v] = for given node, set of <i,j> pairs with X[i,j] fixed to
      value v in {0,1} */
   set <num,num> FIX {NODES, 0..1} init {};
   set <num,num> SUPPORT {NODES} init {};
   num obj {NODES} init .;
   set UNSOLVED   init NODES;
   set CANDIDATES init {};
   set WINNERS    init {};
   num candidate_obj {node in CANDIDATES} = obj[node];
   num min_obj = min(of candidate_obj[*]);
   num argmin_obj;

   num node_this init 1;
   /* fixing to 0 or 1 is equivalent to removing arcs */
   set <str,str> ARCS_THIS =
      (ARCS diff setof {<i,j> in FIX[node_this,0]} <'row'||i,'col'||j>)
      diff setof {<i,j> in FIX[node_this,1], i2 in NSET, j2 in NSET: (i = i2) + (j = j2) = 1}
         <'row'||i2,'col'||j2>;

   /* each iteration of outer loop yields next best solution */
   for {1..k} do;
      /* solve all unsolved nodes concurrently */
      cofor {node in UNSOLVED} do;
         node_this = node;
         put node_this=;
         for {v in 0..1} put FIX[node_this,v]=;
         solve with network /
            LAP direction=directed links=(weight=cost)
            subgraph=(links=ARCS_THIS) out=(assignments=ASSIGNMENTS);
         if card(ASSIGNMENTS) < n then obj[node_this] = .;
         else do;
            obj[node_this] = _OROPTMODEL_num_['OBJECTIVE'];
            SUPPORT[node_this] = {i in NSET, j in NSET: <'row'||i,'col'||j> in ASSIGNMENTS};
         end;
         put obj[node_this] SUPPORT[node_this];
         put ' ';
      end;
      CANDIDATES = CANDIDATES union {node in UNSOLVED: obj[node] ne .};
      UNSOLVED   = {};
      /* find candidate node with best objective */
      for {node in CANDIDATES: obj[node] = min_obj} do;
         argmin_obj = node;
         leave;
      end;
      put min_obj= argmin_obj=;
      put ' ';
      /* record this as next solution */
      numsols = numsols + 1;
      SUPPORT_s[numsols] = SUPPORT[argmin_obj];
      WINNERS            = WINNERS    union {argmin_obj};
      CANDIDATES         = CANDIDATES diff  {argmin_obj};
      /* if supply of good solutions is adequate, can stop branching */
      if numsols = k or
         (card(CANDIDATES) >= k-numsols
         and smallest(k-numsols, of candidate_obj[*]) <= obj[argmin_obj])
      then continue;
      /* create n - 1 new nodes by branching a la Murty
         (condition on first difference) */
      /* r = first difference (cannot be n if square instance) */
      for {r in 1..n-1} do;
         numnodes = numnodes + 1;
         UNSOLVED = UNSOLVED union {numnodes};
         FIX[numnodes,0] = FIX[argmin_obj,0]
            union setof {i in {r},    <(i),j> in SUPPORT[argmin_obj]} <i,j>;
         FIX[numnodes,1] = FIX[argmin_obj,1]
            union setof {i in 1..r-1, <(i),j> in SUPPORT[argmin_obj]} <i,j>;
         /* prune if trivially infeasible */
         if card(FIX[numnodes,0] inter FIX[numnodes,1]) > 0 then do;
            UNSOLVED = UNSOLVED diff {numnodes};
            numnodes = numnodes - 1;
         end;
      end;
   end;

   put WINNERS=;
   for {node in WINNERS} put obj[node] SUPPORT[node];
quit;

